searchData={"items":[{"type":"module","title":"Lorax","doc":"Simple Low-Rank Adaptation (LoRA) implementation","ref":"Lorax.html"},{"type":"module","title":"LoRA model creation - Lorax","doc":"To create a LoRA model, freeze an existing model and inject LoRA layers using `Lorax.inject/2`.\n\n```\nlora_model =\n  model\n  |> Axon.freeze()\n  |> Lorax.inject(%Lorax.Config{\n    r: 2,\n    alpha: 4,\n    dropout: 0.05,\n    target_key: false,\n    target_query: false,\n    target_value: true\n  })\n```\n\nFor more detailed guides, see\n1. [Finetuning LLMs with LoRA](finetuning_gpt_with_lora.livemd)\n2. [Running LLMs with LoRA](running_gpt_with_lora.livemd)\n\n\nLoRA layers are implemented by injecting new nodes into the Axon struct.\nThese LoRA nodes represent the B and A matrices. Each node takes an input `x` and computes `BAx`.\nFurthermore, the LoRA node will receive `Wx` as an input and compute `Wx + BAx`.\nThis isn't the standard implementation, but it simplifies the injection process.","ref":"Lorax.html#module-lora-model-creation"},{"type":"module","title":"Injection Process - Lorax","doc":"Beginning state\n \nflowchart LR\n  A[input id:0] --> B[target id:1]\n \n\nCreate an empty dummy node\n \nflowchart LR\n  A[input id:0] --> B[target id:1] --> C[dummy id:2]\n \n\nCreate lora node with input ids = [0, 2]\n \nflowchart LR\n  A[input id:0] --> B[target id:1] --> C[dummy id:2] --> E[lora id:3]\n  A[input id:0] --> E[lora id:3]\n \n\ntarget takes dummy's id, throw away dummy node\n \nflowchart LR\n  A[input id:0] --> C[target id:2]\n  C[target id:2] --> E[lora id:3]\n  A[input id:0] --> E[lora id:3]\n \n\n\nlora takes target's original id\n \nflowchart LR\n  A[input id:0] --> C[target id:2] --> E[lora id:1]\n  A[input id:0] --> E[lora id:1]\n \n\n\nlora and target are now swapped.\nAny downstream node that relied on node id:1 will now receive `Wx + BAx`","ref":"Lorax.html#module-injection-process"},{"type":"function","title":"Lorax.inject/2","doc":"Returns a modified Axon model with LoRA nodes inserted according to the provided configuration.\n\n`target_key`, `target_query`, `target_value` are required if `target_node_fn` isn't specified","ref":"Lorax.html#inject/2"},{"type":"function","title":"Examples - Lorax.inject/2","doc":"```\nlora_model =\n  model\n  |> Axon.freeze()\n  |> Lorax.inject(%Lorax.Config{\n    r: 2,\n    alpha: 4,\n    dropout: 0.05,\n    target_key: true,\n    target_query: true,\n    target_value: true\n  })\n```","ref":"Lorax.html#inject/2-examples"},{"type":"function","title":"Targeting nodes manually - Lorax.inject/2","doc":"```\nlora_model =\n  model\n  |> Axon.freeze()\n  |> Lorax.inject(%Lorax.Config{\n    r: 2,\n    alpha: 4,\n    dropout: 0.05,\n    target_node_fn: fn %Axon.Node{name: name_fn} ->\n      # names are generated lazily, and look like \"decoder.blocks.11.self_attention.value\"\n      # have to invoke the function to see what layer the node represents\n      # https://github.com/elixir-nx/axon/blob/v0.6.0/lib/axon.ex#L3923\n      name = name_fn.(nil, nil)\n      shortname = String.split(name, \".\") |> List.last()\n\n      if shortname == \"output\" do\n        true\n      else\n        false\n      end\n    end\n  })\n```","ref":"Lorax.html#inject/2-targeting-nodes-manually"},{"type":"module","title":"Lorax.Config","doc":"Config for `Lorax.inject/2`\n\n`r` is the rank in the low-rank matrices used in LoRA.\nA higher value of r increases the expressiveness of the adaptation,\nHowever, it also increases the number of parameters and the computational\ncost. Conversely, a lower value of r makes the adaptation simpler and less\nresource-intensive. Defaults to 1.\n\n`alpha` is a scaling factor that controls the magnitude of changes introduced\nby the low-rank matrices. A higher value of `alpha` means that the\nmodifications made by LoRA have a greater impact on the model's original\nweights. This can lead to more significant changes in the model's behavior.\nA lower value results in more subtle changes. Defaults to 2.\n\n`dropout` specifies the dropout rate applied to the low-rank matrices.\n\n`dropout_seed` determines the seed used for `Nx.Random.key/1` during\ndropout application. When defined, it ensures that the LoRA adapter\nproduces consistent tensor values, assuming that other layers also have\ndeterministic outputs.\n\n`param_type` specifies the numerical representation for the A and B\nmatrices. Defaults to float32\n\n`target_query` specifies whether to apply LoRA to all query matrices in an\nattention block. Defaults to true.\n\n`target_value` specifies whether to apply LoRA to all value matrices in an\nattention block. Defaults to true.\n\n`target_key` specifies whether to apply LoRA to all key matrices in an\nattention block. Defaults to true.","ref":"Lorax.Config.html"},{"type":"module","title":"Lorax.Lcm","doc":"","ref":"Lorax.Lcm.html"},{"type":"function","title":"Lorax.Lcm.load!/1","doc":"Loads some serialized LCM LoRA data and returns a returns a tuple of `{%Lorax.Config{}, lora_params}`","ref":"Lorax.Lcm.html#load!/1"},{"type":"function","title":"Lorax.Lcm.read!/1","doc":"Reads some LCM LoRA safetensor file and returns a tuple of `{%Lorax.Config{}, lora_params}`","ref":"Lorax.Lcm.html#read!/1"},{"type":"module","title":"Lorax.Params","doc":"Helper module for loading, downloading, filtering, and calculating the size of Axon parameters","ref":"Lorax.Params.html"},{"type":"function","title":"Lorax.Params.file_load!/1","doc":"Loads parameters from file path","ref":"Lorax.Params.html#file_load!/1"},{"type":"function","title":"Lorax.Params.filter/2","doc":"Returns LoRA only params from a merged param map","ref":"Lorax.Params.html#filter/2"},{"type":"function","title":"Lorax.Params.kino_download/3","doc":"Creates a Kino widget for downloading params map.\nMust be placed in the last line of a Livebook cell.","ref":"Lorax.Params.html#kino_download/3"},{"type":"function","title":"Lorax.Params.kino_file_load!/1","doc":"Creates a Kino widget for uploading serialized params file.\nMust be placed in the last line of a Livebook cell.","ref":"Lorax.Params.html#kino_file_load!/1"},{"type":"function","title":"Lorax.Params.size/1","doc":"Calculates total bytes of all the tensors inside a parameter map","ref":"Lorax.Params.html#size/1"},{"type":"module","title":"Lorax.Shape","doc":"","ref":"Lorax.Shape.html"},{"type":"function","title":"Lorax.Shape.calc_ab/3","doc":"Infers the shape of the LoRA matrices. Only supports :dense and :conv operations\n\n#","ref":"Lorax.Shape.html#calc_ab/3"},{"type":"function","title":"Dense LoRA kernels - Lorax.Shape.calc_ab/3","doc":"Suppose we have a target node `W` and input tensor `x`,\nDuring injection, the A matrix will project the input tensor down to\nr-dimensional space. Afterwards, the B matrix will project the result back up\nto some unknown dimensionality. To figure out the output dimensionality,\nwe inspect W's kernel shape.\n\n#","ref":"Lorax.Shape.html#calc_ab/3-dense-lora-kernels"},{"type":"function","title":"Convolution LoRA kernels - Lorax.Shape.calc_ab/3","doc":"In addition to figuring out the output dimensionality, we need to retrieve\nthe convolution kernel_size of `W`. When calling Axon.conv, it's passed as an\noption, but then is no longer stored inside the Axon node. To figure out the\nkernel size, we also inspect W's kernel shape.","ref":"Lorax.Shape.html#calc_ab/3-convolution-lora-kernels"},{"type":"extras","title":"Lorax","doc":"# Lorax\n\n[![Module Version](https://img.shields.io/hexpm/v/lorax.svg)](https://hex.pm/packages/lorax)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-lightgreen.svg)](https://hexdocs.pm/lorax/)\n\nThis package implements [Low-Rank Adaptation (LoRA)](https://arxiv.org/abs/2106.09685), a popular method for fine-tuning large language models.","ref":"readme.html"},{"type":"extras","title":"Installation - Lorax","doc":"This package can be installed by adding `lorax` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:lorax, \"~> 0.1.0\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Fine-tuning an LLM with LoRA - Lorax","doc":"In general,\n\n1. Import your model\n2. Inject trainable LoRA parameters\n3. Train LoRA model\n4. Download LoRA only params\n\n```\n{:ok, model_info} = Bumblebee.load_model({:hf, \"gpt2\"})\n%{model: gpt2_model, params: gpt2_params} = model_info\n\nlora_model =\n  gpt2_model\n  |> Axon.freeze()\n  |> Lorax.inject(%Lorax.Config{\n      r: 2,\n      alpha: 4,\n      dropout: 0.05\n  })\n\nlora_merged_params =\n  Axon.build(lora_model, mode: :train)\n  |> Axon.Loop.trainer(custom_loss_fn, Polaris.Optimizers.adam(learning_rate: 3.0e-4))\n  |> Axon.Loop.run(train_batch_stream, gpt2_params, epochs: 3, iterations: 1000, compiler: EXLA)\n\nlora_params = lora_merged_params\n  |> Lorax.Params.filter(gpt2_params)\n  |> Lorax.Params.kino_download()\n```\n\nIn practice, every model has some unique architecture that you need to account for.\nFor more detailed guides, see\n\n1. [Finetuning LLMs with LoRA](guides/finetuning_gpt_with_lora.livemd)\n2. [Running LLMs with LoRA](guides/running_gpt_with_lora.livemd)","ref":"readme.html#fine-tuning-an-llm-with-lora"},{"type":"extras","title":"Default Settings - Lorax","doc":"The default config applies LoRA to all query, key, value matrices. r = 1, alpha = 2.\n\nThe LoRA paper demonstrated that adapting only the query and value matrices with r = 1 achieved effective fine-tuning results. However, for larger language models, people often choose much higher values of r and sometimes target all linear layers.","ref":"readme.html#default-settings"},{"type":"extras","title":"Recommended Settings - Lorax","doc":"These settings works well for fine-tuning smaller LLMs (~1b param models)\n\n```\nLora Config\n- r value  = at least 2\n- alpha value = r * 2\n\nTraining\n- learning_rate of 3.0e-4 with an adam optimizer\n\nText Generation\n- multinomial sampling\n- p = 0.06 or 0.08 for more variety (or if you experience repetitive results)\n```\n\nFor more details on configuring LoRA hyperparameters, see this [post](https://lightning.ai/pages/community/lora-insights/) by Sebastian Raschka.","ref":"readme.html#recommended-settings"},{"type":"extras","title":"Limitations - Lorax","doc":"1. GPU Memory Requirements: Although LoRA reduces the GPU requirements for fine-tuning, larger LLMs beyond GPT2 still demand GPUs with substantial vRAM. Inadequate memory management can lead to cuda OOM crashes.\n\n2. Fine-Tuning Speed: The training speed of this library isn't on par with Huggingface's PEFT library. Further optimizations can be done to close the gap.\n\nNote: For minor fine-tuning tasks without a GPU, the BinaryBackend is a viable option, often resulting in smoother training runs. Future updates will focus on minimizing GPU memory usage by reducing the amount of tensors stored during training, and potentially a QLoRA implementation one day.","ref":"readme.html#limitations"},{"type":"extras","title":"Finetuning LLMs with LoRA","doc":"# Finetuning LLMs with LoRA\n\n```elixir\nMix.install([\n  {:bumblebee, \"~> 0.5.3\"},\n  {:axon, \"~> 0.6.1\"},\n  {:nx, \"~> 0.7.1\"},\n  {:exla, \"~> 0.7.1\"},\n  {:lorax, \"~> 0.2.1\"},\n  {:req, \"~> 0.4.0\"},\n  {:kino, \"~> 0.12.3\"}\n])\n\nNx.default_backend(EXLA.Backend)\n```","ref":"finetuning_gpt_with_lora.html"},{"type":"extras","title":"Introduction - Finetuning LLMs with LoRA","doc":"This notebook will show how to train a LoRA adapter w/ GPT2. Most of this notebook is copied from https://hexdocs.pm/bumblebee/fine_tuning.html. If you want to learn more about how this training setup works, check that livebook instead. The LoRA specific details are further down below.","ref":"finetuning_gpt_with_lora.html#introduction"},{"type":"extras","title":"Hyperparameters - Finetuning LLMs with LoRA","doc":"```elixir\nbatch_size = 2\nsequence_length = 256\nr = 2\nlora_alpha = 4\nlora_dropout = 0.05\n\n:ok\n```\n\n \n\n```\n:ok\n```","ref":"finetuning_gpt_with_lora.html#hyperparameters"},{"type":"extras","title":"Load a model - Finetuning LLMs with LoRA","doc":"```elixir\n{:ok, spec} = Bumblebee.load_spec({:hf, \"gpt2\"})\n{:ok, model} = Bumblebee.load_model({:hf, \"gpt2\"}, spec: spec)\n{:ok, tokenizer} = Bumblebee.load_tokenizer({:hf, \"gpt2\"})\n{:ok, generation_config} = Bumblebee.load_generation_config({:hf, \"gpt2\"})\n\n:ok\n```\n\n \n\n```\n\n17:24:05.365 [info] TfrtCpuClient created.\n\n```\n\n \n\n```\n:ok\n```","ref":"finetuning_gpt_with_lora.html#load-a-model"},{"type":"extras","title":"Prepare a dataset - Finetuning LLMs with LoRA","doc":"We'll be using some data scraped from Elixirforum.\nIt's structured like this.\n\n```\n Ideas for an Open Source Discord \n\n WildYorkies \n\nI remember seeing on the Elixir ...\n 3 likes \n```\n\n```elixir\ntext =\n  Req.get!(\"https://raw.githubusercontent.com/wtedw/lorax/main/data/elixir-discussion.txt\").body\n\n:ok\n```\n\n \n\n```\n:ok\n```\n\n```elixir\ntokenized_text = %{\"input_ids\" => input_ids} = Bumblebee.apply_tokenizer(tokenizer, text)\nn_tokens = Nx.size(input_ids)\nn_train = round(n_tokens * 0.9)\nn_val = n_tokens - n_train\n\ntrain_data =\n  for {input_key, tokenized_values}  \n\n```\n%{\n  \"attention_mask\" => #Nx.Tensor \n    [\n      [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, ...]\n    ]\n  >,\n  \"input_ids\" => #Nx.Tensor \n    [\n      [952, 30120, 25, 198, 40, 447, 247, 76, 2111, 284, 2050, 922, 7572, 329, 2615, 42652, 9643, 6725, 13, 198, 40, 481, 467, 4622, 503, 286, 7243, 994, 13, 1002, 345, 447, 247, 260, 655, 3599, 503, 314, 561, 1950, 326, 345, 467, 329, 262, 10314, 26, 42652, ...]\n    ]\n  >,\n  \"token_type_ids\" => #Nx.Tensor \n    [\n      [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, ...]\n    ]\n  >\n}\n```\n\n```elixir\ndefmodule DataStream do\n  def get_batch_stream(%{\"input_ids\" => input_ids} = data, batch_size, block_size, opts \\\\ []) do\n    seed = Keyword.get(opts, :seed, 1337)\n\n    Stream.resource(\n      # initialization function\n      fn ->\n        Nx.Random.key(seed)\n      end,\n      # generation function\n      fn key ->\n        {_b, t} = Nx.shape(input_ids)\n\n        data =\n          for {k, v}   Enum.map(fn i -> Nx.slice_along_axis(tensor, i, block_size, axis: -1) end)\n              |> Nx.stack()\n\n            {k, batch_slice}\n          end\n\n        # y represents all the predicted next tokens (input_ids shifted by 1)\n        y =\n          ix\n          |> Enum.map(fn i ->\n            data[\"input_ids\"] |> Nx.slice_along_axis(i + 1, block_size, axis: -1)\n          end)\n          |> Nx.stack()\n          |> Nx.flatten()\n\n        out_data = {x, y}\n\n        {[out_data], new_key}\n      end,\n      fn _ -> :ok end\n    )\n  end\nend\n```\n\n \n\n```\n{:module, DataStream,  >, {:get_batch_stream, 4}}\n```\n\nYou can see what a single batch looks like by grabbing 1 from the stream:\n\n```elixir\ntrain_batch_stream = DataStream.get_batch_stream(train_data, batch_size, sequence_length)\ntest_batch_stream = DataStream.get_batch_stream(test_data, batch_size, sequence_length)\n\n[{x, y}] = train_batch_stream |> Enum.take(1)\n[{x_val, y_val}] = test_batch_stream |> Enum.take(1)\n\nBumblebee.Tokenizer.decode(tokenizer, x[\"input_ids\"]) |> IO.inspect()\nIO.puts(\"=====\")\nBumblebee.Tokenizer.decode(tokenizer, y) |> IO.inspect()\n\n:ok\n```\n\n \n\n```\n[\" to how Greg’s Event Store works with its competing consumers model. Why would you want to do this? It allows handlers to run at different speeds, typically you have slow async handlers that can lag behind (e.g. sending emails, third party API requets). But you don’t want them to hold up read model projections to minimise query latency.\\nAutonomous subscriptions allows you to add new handlers and replay all events from the beginning of time, or restart a handler to rebuild a projection. I’ve implemented a hybrid push/pull model for the Event Store subscriptions where appended events are published to subscribers, but they are buffered per subscriber and use back-pressure to ensure the subscriber isn’t overwhelmed. The subscription falls back to pulling events from the store when it gets too far behind, until caught up again.\\nYou could use GenStage for this, but I would recommend using an individual flow pipeline per handler; not one flow for all handlers. Since GenStage's broadcast dispatcher can only go as fast as the slowest consumer. You also want to have any event handlers run from the event store, after the events have been atomically persisted. Appending events to the store should guarantee that a success reply is returned\",\n \" by a more fundamental principle - of developer happiness and a system that ‘just makes sense’. Of course the syntax is itself inspired by this, but it goes beyond syntax.\\nHaving said that, I think José has also tried to stay true to Erlang and this certainly shows when using Elixir.\\nWhen would I use Ruby? When I don’t need Elixir When I write a script for the server, or need to put a site up quick, or have a smaller project in mind I would use Ruby. Mainly for two reasons: I know it, and there is a huge community/set of libraries out there. Chances are if you want to do something someone already has in Ruby.\\nThat may well change as I learn Elixir. I’m hoping it does actually, as my brain can’t hold too much information so sticking to one language would be preferential for me\\n\\n 2 likes \\n\\n gnat \\n\\nAstonJ:\\nmy brain can’t hold too much information so sticking to one language would be preferential for me\\nThat was at least part of where I was coming from in raising the original question. I used to code\"]\n=====\n\" how Greg’s Event Store works with its competing consumers model. Why would you want to do this? It allows handlers to run at different speeds, typically you have slow async handlers that can lag behind (e.g. sending emails, third party API requets). But you don’t want them to hold up read model projections to minimise query latency.\\nAutonomous subscriptions allows you to add new handlers and replay all events from the beginning of time, or restart a handler to rebuild a projection. I’ve implemented a hybrid push/pull model for the Event Store subscriptions where appended events are published to subscribers, but they are buffered per subscriber and use back-pressure to ensure the subscriber isn’t overwhelmed. The subscription falls back to pulling events from the store when it gets too far behind, until caught up again.\\nYou could use GenStage for this, but I would recommend using an individual flow pipeline per handler; not one flow for all handlers. Since GenStage's broadcast dispatcher can only go as fast as the slowest consumer. You also want to have any event handlers run from the event store, after the events have been atomically persisted. Appending events to the store should guarantee that a success reply is returned only a more fundamental principle - of developer happiness and a system that ‘just makes sense’. Of course the syntax is itself inspired by this, but it goes beyond syntax.\\nHaving said that, I think José has also tried to stay true to Erlang and this certainly shows when using Elixir.\\nWhen would I use Ruby? When I don’t need Elixir When I write a script for the server, or need to put a site up quick, or have a smaller project in mind I would use Ruby. Mainly for two reasons: I know it, and there is a huge community/set of libraries out there. Chances are if you want to do something someone already has in Ruby.\\nThat may well change as I learn Elixir. I’m hoping it does actually, as my brain can’t hold too much information so sticking to one language would be preferential for me\\n\\n 2 likes \\n\\n gnat \\n\\nAstonJ:\\nmy brain can’t hold too much information so sticking to one language would be preferential for me\\nThat was at least part of where I was coming from in raising the original question. I used to code pretty\"\n```\n\n \n\n```\n:ok\n```","ref":"finetuning_gpt_with_lora.html#prepare-a-dataset"},{"type":"extras","title":"Train the model - Finetuning LLMs with LoRA","doc":"```elixir\n%{model: model, params: params} = model\n\nmodel\n```\n\n \n\n```\n#Axon \n```\n\n```elixir\n[{input, _}] = Enum.take(train_batch_stream, 1)\nAxon.get_output_shape(model, input)\n```\n\n \n\n```\n%{\n  cache: #Axon.None ,\n  hidden_states: #Axon.None ,\n  attentions: #Axon.None ,\n  cross_attentions: #Axon.None ,\n  logits: {2, 256, 50257}\n}\n```\n\nWe'll need to freeze the original parameters in our model before injecting. That way we can train only the LoRA nodes.\n\n```elixir\nlora_model =\n  model\n  |> Axon.freeze()\n  |> Lorax.inject(%Lorax.Config{\n    r: r,\n    alpha: lora_alpha,\n    dropout: lora_dropout,\n    target_key: true,\n    target_query: true,\n    target_value: true\n  })\n```\n\n \n\n```\n#Axon \n```\n\nWe'll reshape the usual GPT2 output so that we can use it with `categorical_cross_entropy`\n\n```elixir\ndefmodule CommonTrain do\n  import Nx.Defn\n\n  defn custom_predict_fn(model_predict_fn, params, input) do\n    %{prediction: preds} = out = model_predict_fn.(params, input)\n\n    # Output of GPT2 model is a map containing logits and other tensors\n    logits = preds.logits\n\n    {b, t, c} = Nx.shape(logits)\n    reshaped = Nx.reshape(logits, {b * t, c})\n    %{out | prediction: reshaped}\n  end\n\n  def custom_loss_fn(y_true, y_pred) do\n    Axon.Losses.categorical_cross_entropy(y_true, y_pred,\n      from_logits: true,\n      sparse: true,\n      reduction: :mean\n    )\n  end\nend\n\n{init_fn, predict_fn} = Axon.build(lora_model, mode: :train)\ncustom_predict_fn = &CommonTrain.custom_predict_fn(predict_fn, &1, &2)\ncustom_loss_fn = &CommonTrain.custom_loss_fn(&1, &2)\n\nlora_merged_params =\n  {init_fn, custom_predict_fn}\n  |> Axon.Loop.trainer(custom_loss_fn, Polaris.Optimizers.adam(learning_rate: 3.0e-4))\n  |> Axon.Loop.run(train_batch_stream, params, epochs: 1, iterations: 300, compiler: EXLA)\n\n:ok\n```\n\n \n\n```\n\n17:42:55.140 [debug] Forwarding options: [compiler: EXLA] to JIT compiler\nEpoch: 0, Batch: 250, loss: 3.6697373\n```\n\n \n\n```\n:ok\n```","ref":"finetuning_gpt_with_lora.html#train-the-model"},{"type":"extras","title":"Download LoRA params - Finetuning LLMs with LoRA","doc":"When training, Axon will return all the parameters needed to run the model. If you want to download the LoRA-only parameters, you'll need to filter for them first before downloading.\n\n```elixir\n# Method #1\nlora_only =\n  lora_merged_params\n  |> Lorax.Params.filter(params)\n  |> Lorax.Params.kino_download()\n\n# Method #2\n# serialized =\n#   Lorax.Params.filter(lora_merged_params, params)\n#   |> Nx.serialize()\n\n# File.write!(\"/ /test.lorax\", serialized)\n```","ref":"finetuning_gpt_with_lora.html#download-lora-params"},{"type":"extras","title":"Testing out text Generation - Finetuning LLMs with LoRA","doc":"```elixir\nlora_model_info = %{model: lora_model, params: lora_merged_params, spec: spec}\n\nlora_generation_config =\n  Bumblebee.configure(generation_config,\n    max_new_tokens: 256,\n    strategy: %{type: :multinomial_sampling, top_p: 0.8}\n  )\n\nserving =\n  Bumblebee.Text.generation(lora_model_info, tokenizer, lora_generation_config,\n    compile: [batch_size: 1, sequence_length: 256],\n    stream: true,\n    defn_options: [compiler: EXLA, lazy_transfers: :always]\n  )\n\nKino.start_child({Nx.Serving, name: Llama, serving: serving})\n```\n\n \n\n```\n{:ok, #PID }\n```\n\nWe'll kickstart the text generation using some ` text ` string. This is how the training data is formatted and will trigger the LoRA neurons to activate. With only 300 iterations, the model has already learned how to output text similar to our training data\n\n```elixir\nNx.Serving.batched_run(Llama, \" Elixir 2.0 released\") |> Enum.each(&IO.write/1)\n```\n\n \n\n```\n \n\n 0 likes \n\n jlarperendan \n\nThe elixir.clipper gem is really very elegant, you can even use it with other compilers if you don't need it in your code base. It really helps here!\n ilax13 \n\nThe name of the library is good! It already provides a lot of benefit from Elixir, so I guess it's useful for intermediate project developers to have one with both Elixir and Elixir 2.0.1.\n\n 0 likes \n\n benaeutz \n\nElixir 2.0 is really powerful. Even though there's some backend changes, Elixir 2.0 is one of the few that changes the behavior of main mode and does not change runtime dependencies. So, if you have multiple compilers that can add the language layer or break the runtime dependencies, then the big advantage of Elixir is the compiler ecosystem. In many cases, the exception to this is compiler breaking.\n 1 likes \n\n avnicom 1 likes\n```\n\n \n\n```\n:ok\n```","ref":"finetuning_gpt_with_lora.html#testing-out-text-generation"},{"type":"extras","title":"Running LLMs with LoRA","doc":"# Running LLMs with LoRA\n\n```elixir\nMix.install([\n  {:bumblebee, \"~> 0.5.3\"},\n  {:axon, \"~> 0.6.1\"},\n  {:nx, \"~> 0.7.1\"},\n  {:exla, \"~> 0.7.1\"},\n  {:explorer, \"~> 0.7.0\"},\n  {:lorax, \"~> 0.2.1\"},\n  {:req, \"~> 0.4.0\"},\n  {:kino, \"~> 0.12.3\"}\n])\n\nNx.default_backend(EXLA.Backend)\n```","ref":"running_gpt_with_lora.html"},{"type":"extras","title":"Introduction - Running LLMs with LoRA","doc":"This notebook demonstrates how to run a text-generating model like GPT2 with LoRA. The basic steps:\n\n1. Define the model as it was during fine-tuning\n2. Merge the LoRA parameters with the base model weights.\n3. Run the model\n\nWe'll be using a LoRA file that was trained on the Elixir Chat/Discussion section. The data can be found [here](https://github.com/wtedw/lorax/blob/main/data/elixir-discussion.txt).","ref":"running_gpt_with_lora.html#introduction"},{"type":"extras","title":"Load original model - Running LLMs with LoRA","doc":"```elixir\n{:ok, spec} = Bumblebee.load_spec({:hf, \"gpt2\"})\n{:ok, model} = Bumblebee.load_model({:hf, \"gpt2\"}, spec: spec)\n{:ok, tokenizer} = Bumblebee.load_tokenizer({:hf, \"gpt2\"})\n{:ok, generation_config} = Bumblebee.load_generation_config({:hf, \"gpt2\"})\n\n%{model: model, params: gpt2_params} = model\n\n:ok\n```\n\n \n\n```\n:ok\n```","ref":"running_gpt_with_lora.html#load-original-model"},{"type":"extras","title":"Define LoRA Model - Running LLMs with LoRA","doc":"Inject LoRA layers as it was during fine-tuning to ensure correct model compilation.\n\n```elixir\nr = 4\nlora_alpha = 8\nlora_dropout = 0.05\n\nlora_model =\n  model\n  |> Axon.freeze()\n  |> Lorax.inject(%Lorax.Config{\n    r: r,\n    alpha: lora_alpha,\n    dropout: lora_dropout,\n    target_key: true,\n    target_query: true,\n    target_value: true\n  })\n```\n\n \n\n```\n#Axon \n```","ref":"running_gpt_with_lora.html#define-lora-model"},{"type":"extras","title":"Load LoRA Params - Running LLMs with LoRA","doc":"There's 3 main ways to load your LoRA params\n\n1. URL: Use whatever HTTP library to retrieve the binary, and Nx.deserialize to get the map of tensor values\n2. File: Similar flow as URL. use File.read to retrieve the binary, Nx.deserialize. For convenience, you can use Lorax.Params.file_load!()\n3. Kino: GUI file picker. This requires a two-step process. You need to have one Kino input cell, and another cell to read the file input.\n\nSee the code below for examples\n\n```elixir\n# Method 1: URL downlaod\nlora_serialized =\n  Req.get!(\"https://raw.githubusercontent.com/wtedw/lorax/main/params/elixir-chat-r4a8.lorax\").body\nlora_only_params = Nx.deserialize(lora_serialized)\nlora_only_params |> Map.keys()\n\n# Method 2: File download\n# lora_only_params = Lorax.Params.file_load!(\" \")\n\n# Method 3: Kino\n# cell #1\n# input = Kino.Input.file(\"Lorax Params\")\n#\n# cell #2\n# lora_only_params = Lorax.Params.kino_file_load!(input)\n# merged_params = Lorax.Params.merge_params(lora_only_params, gpt2_params)\n```\n\n \n\n```\n[\"dropout_28\", \"decoder.blocks.5.self_attention_dropout\", \"lora_11\", \"lora_31\", \"lora_10\",\n \"decoder.blocks.11.self_attention_dropout\", \"dropout_9\", \"lora_28\", \"dropout_25\",\n \"decoder.blocks.1.self_attention_dropout\", \"lora_13\", \"lora_20\", \"lora_7\", \"dropout_6\", \"lora_19\",\n \"decoder.blocks.2.self_attention_dropout\", \"lora_5\", \"lora_30\",\n \"decoder.blocks.7.self_attention_dropout\", \"lora_2\", \"lora_24\", \"dropout_1\", \"lora_6\",\n \"dropout_10\", \"dropout_3\", \"dropout_15\", \"lora_27\", \"dropout_7\", \"lora_35\", \"lora_22\",\n \"dropout_21\", \"decoder.blocks.3.self_attention_dropout\", \"lora_0\", \"dropout_27\", \"dropout_24\",\n \"lora_9\", \"decoder.blocks.4.self_attention_dropout\", \"lora_32\", \"dropout_34\",\n \"decoder.blocks.10.self_attention_dropout\", \"lora_23\", \"dropout_0\",\n \"decoder.blocks.6.self_attention_dropout\", \"lora_12\", \"dropout_16\", \"lora_34\", \"dropout_19\",\n \"lora_18\", \"lora_3\", \"lora_17\", ...]\n```","ref":"running_gpt_with_lora.html#load-lora-params"},{"type":"extras","title":"Merge Params - Running LLMs with LoRA","doc":"Axon expects one single map of all the parameter values. So although we've loaded the LoRA params, we need to merge them with the original parameters.\n\n```elixir\nmerged_params = Map.merge(gpt2_params, lora_only_params)\n\n:ok\n```\n\n \n\n```\n:ok\n```","ref":"running_gpt_with_lora.html#merge-params"},{"type":"extras","title":"Inference Prepwork - Running LLMs with LoRA","doc":"The sampling method used here is non-deterministic, so the output is different everytime you run this cell + the bottom cell. Using top_p value of 0.6 may generate more coherent sentences, but tends to repeat. 0.7 and 0.8 is a good sweet spot.\n\n```elixir\nlora_model_info = %{model: lora_model, params: merged_params, spec: spec}\n\nlora_generation_config =\n  Bumblebee.configure(generation_config,\n    max_new_tokens: 512,\n    strategy: %{type: :multinomial_sampling, top_p: 0.7}\n  )\n\nserving =\n  Bumblebee.Text.generation(lora_model_info, tokenizer, lora_generation_config,\n    compile: [batch_size: 1, sequence_length: 512],\n    stream: true,\n    defn_options: [compiler: EXLA, lazy_transfers: :always]\n  )\n\nKino.start_child({Nx.Serving, name: Llama, serving: serving})\n```\n\n \n\n```\n{:ok, #PID }\n```","ref":"running_gpt_with_lora.html#inference-prepwork"},{"type":"extras","title":"Inference Results - Running LLMs with LoRA","doc":"The training data is formatted like this:\n\n```\n Ideas for an Open Source Discord \n\n WildYorkies \n\nI remember seeing on the Elixir subreddit [...]\n\n 3 likes \n```\n\nYou can simulate a thread by kickstarting the text generation with some ` text ` string, and the LLM will give a prediction of what the general Elixir community will say.\n\n```elixir\nNx.Serving.batched_run(\n  Llama,\n  \" Elixir 2.0 is released! New features include \"\n)\n|> Enum.each(&IO.write/1)\n```\n\n \n\n```\n\n\n xjalasz \n\nElixir 2.0 is released! New features include\nThis means that you can now deploy your Elixir 2 projects without having to use a tool like docker.\n\n 1 like \n\n jake \n\nAs always, thanks for the help.\nI am on a 10 day cruise to Toronto in July with the goal of finishing up Elixir 2.0 in less than one week.\n\n 1 like \n\n pianotato \n\nThanks for all the support!\n\n 0 likes \n\n thekali \n\nThanks for your time and effort.\n\n 0 likes \n\n mangai \n\nThank you! I have no idea how long you have been on this blog.\n\n 1 like \n\n juicy_champion \n\nThanks!\n\n 0 likes \n\n haha 1 like \n\n thompson3 \n\nThank you for the support!\n\n 0 likes \n\n phrobotics \n\nThanks for your time and effort.\n\n 0 likes \n\n yelp \n\nThanks!\n\n 0 likes \n\n phrobotics \n\nThanks!\n\n 0 likes \n\n jake \n\nThanks for your time and effort.\n\n 0 likes \n\n jake \n\nThanks!\n\n 0 likes \n\n juicy_champion \n\nThanks!\n\n 0 likes \n\n juicy_champion \n\nThanks!\n\n 0 likes \n\n gw\n```\n\n \n\n```\n:ok\n```","ref":"running_gpt_with_lora.html#inference-results"},{"type":"extras","title":"Stable Diffusion w/ LCM LoRA","doc":"# Stable Diffusion w/ LCM LoRA\n\n```elixir\nMix.install(\n  [\n    {:bumblebee, git: \"https://github.com/elixir-nx/bumblebee.git\"},\n    {:nx, \"~> 0.6.1\", override: true},\n    {:exla, \"~> 0.6.1\"},\n    {:kino, \"~> 0.11.0\"},\n    {:lorax, git: \"https://github.com/wtedw/lorax.git\"},\n    {:axon, [env: :prod, git: \"https://github.com/elixir-nx/axon.git\", override: true]},\n    {:req, \"~> 0.4.0\"}\n  ],\n  config: [nx: [default_backend: EXLA.Backend]]\n)\n\nNx.global_default_backend(EXLA.Backend)\n```","ref":"lcm_lora.html"},{"type":"extras","title":"Load SD1.5, LCM LoRA, and LCM scheduler - Stable Diffusion w/ LCM LoRA","doc":"LCM LoRA requires two key things\n\n1. Loading the LoRA adapter for the appropriate Stable Diffusion unet model. We'll be using the LoRA file for SD 1.5\n2. LCMScheduler. The adapter will not work if any other scheduler is used.\n\nWe'll first download the safetensors from HF and call the Lorax library to convert the parameters to something Axon can use. In addition to converting the params, Lorax will provide a config that describes how new parameters should be injected into the Axon model.\n\n```elixir\nrepo_id = \"runwayml/stable-diffusion-v1-5\"\nopts = [params_variant: \"fp16\"]\n\n{:ok, tokenizer} = Bumblebee.load_tokenizer({:hf, \"openai/clip-vit-large-patch14\"})\n{:ok, clip} = Bumblebee.load_model({:hf, repo_id, subdir: \"text_encoder\"}, opts)\n{:ok, unet} = Bumblebee.load_model({:hf, repo_id, subdir: \"unet\"}, opts)\n{:ok, vae} = Bumblebee.load_model({:hf, repo_id, subdir: \"vae\"}, [architecture: :decoder] ++ opts)\n{:ok, scheduler} = Bumblebee.load_scheduler({:hf, repo_id, subdir: \"scheduler\"})\n{:ok, featurizer} = Bumblebee.load_featurizer({:hf, repo_id, subdir: \"feature_extractor\"})\n{:ok, safety_checker} = Bumblebee.load_model({:hf, repo_id, subdir: \"safety_checker\"}, opts)\n\n# Option #1, Download LCM LoRA from HF\nresp =\n  Req.get!(\n    \"https://huggingface.co/latent-consistency/lcm-lora-sdv1-5/resolve/main/pytorch_lora_weights.safetensors?download=true\"\n  )\n\nparam_data = resp.body\n{config, lcm_lora_params} = Lorax.Lcm.load!(param_data)\n\n# Option #2, Load Locally\n# {config, lcm_lora_params} =\n#   Lorax.Lcm.read!(\"/Users/[user]/.../pytorch_lora_weights.safetensors\")\n\n# Axon expects one map containing all the layer names -> tensors. \n# We'll merge the LCM params with the original SD params.\nlcm_unet = %{\n  unet\n  | model: Lorax.inject(unet.model, config),\n    params: Map.merge(lcm_lora_params, unet.params)\n}\n\nlcm_scheduler = %Bumblebee.Diffusion.LcmScheduler{}\n```","ref":"lcm_lora.html#load-sd1-5-lcm-lora-and-lcm-scheduler"},{"type":"extras","title":"Create serving + Kino prompts - Stable Diffusion w/ LCM LoRA","doc":"```elixir\nserving =\n  Bumblebee.Diffusion.StableDiffusion.text_to_image(clip, lcm_unet, vae, tokenizer, lcm_scheduler,\n    num_steps: 4,\n    num_images_per_prompt: 1,\n    safety_checker_featurizer: featurizer,\n    guidance_scale: 1.0,\n    compile: [batch_size: 1, sequence_length: 60],\n    defn_options: [compiler: EXLA]\n  )\n\n# compare this with regular SD 1.5\n# serving =\n#   Bumblebee.Diffusion.StableDiffusion.text_to_image(clip, unet, vae, tokenizer, scheduler,\n#     num_steps: 4,\n#     num_images_per_prompt: 1,\n#     safety_checker_featurizer: featurizer,\n#     compile: [batch_size: 1, sequence_length: 60],\n#     defn_options: [compiler: EXLA]\n#   )\n\nprompt_input =\n  Kino.Input.text(\"Prompt\",\n    default: \"astronaut in the desert, high quality, detailed\"\n  )\n\nnegative_prompt_input = Kino.Input.text(\"Negative Prompt\", default: \"blurry\")\n\nKino.Layout.grid([prompt_input, negative_prompt_input])\n```\n\n#","ref":"lcm_lora.html#create-serving-kino-prompts"},{"type":"extras","title":"Text to Image - Stable Diffusion w/ LCM LoRA","doc":"some notes about generation\n\n* Although LCM LoRA speeds up image generation, it's sensitive to \"correct\" prompts.\n* Some random seeds produce good images, some are complete gibberish (especially if you only have one word like \"turtle\" as your prompt).\n* The LCM folks say 2-8 steps is a good range, but I found that 4 steps and guidance_scale = 1.0 is best. 8 steps causes too much denoising to happen, so the images appear more saturated.\n\n \n\nWe are ready to generate images!\n\n```elixir\nprompt = Kino.Input.read(prompt_input)\nnegative_prompt = Kino.Input.read(negative_prompt_input)\noutput = Nx.Serving.run(serving, %{prompt: prompt, negative_prompt: negative_prompt})\n\nfor result   Kino.Layout.grid(columns: 2)\n```","ref":"lcm_lora.html#text-to-image"},{"type":"extras","title":"Stable Diffusion w/ LCM (Realtime)","doc":"# Stable Diffusion w/ LCM (Realtime)\n\n```elixir\nMix.install(\n  [\n    {:bumblebee, git: \"https://github.com/elixir-nx/bumblebee.git\"},\n    {:nx, \"~> 0.6.1\", override: true},\n    {:exla, \"~> 0.6.1\"},\n    {:kino, \"~> 0.11.0\"},\n    {:lorax, git: \"https://github.com/wtedw/lorax.git\"},\n    {:axon, [env: :prod, git: \"https://github.com/elixir-nx/axon.git\", override: true]},\n    {:req, \"~> 0.4.0\"}\n  ],\n  config: [nx: [default_backend: EXLA.Backend]]\n)\n\nNx.global_default_backend(EXLA.Backend)\n```","ref":"lcm_lora_realtime.html"},{"type":"extras","title":"Load SD1.5, LCM LoRA, and LCM scheduler - Stable Diffusion w/ LCM (Realtime)","doc":"LCM LoRA requires two key things\n\n1. Loading the LoRA adapter for the appropriate Stable Diffusion unet model. We'll be using the LoRA file for SD 1.5\n2. LCMScheduler. The adapter will not work if any other scheduler is used.\n\nWe'll first download the safetensors from HF and call the Lorax library to convert the parameters to something Axon can use. In addition to converting the params, Lorax will provide a config that describes how new parameters should be injected into the Axon model.\n\n```elixir\nrepo_id = \"runwayml/stable-diffusion-v1-5\"\nopts = [params_variant: \"fp16\"]\n\n{:ok, tokenizer} = Bumblebee.load_tokenizer({:hf, \"openai/clip-vit-large-patch14\"})\n{:ok, clip} = Bumblebee.load_model({:hf, repo_id, subdir: \"text_encoder\"}, opts)\n{:ok, unet} = Bumblebee.load_model({:hf, repo_id, subdir: \"unet\"}, opts)\n{:ok, vae} = Bumblebee.load_model({:hf, repo_id, subdir: \"vae\"}, [architecture: :decoder] ++ opts)\n{:ok, scheduler} = Bumblebee.load_scheduler({:hf, repo_id, subdir: \"scheduler\"})\n{:ok, featurizer} = Bumblebee.load_featurizer({:hf, repo_id, subdir: \"feature_extractor\"})\n{:ok, safety_checker} = Bumblebee.load_model({:hf, repo_id, subdir: \"safety_checker\"}, opts)\n\n# Option #1, Download LCM LoRA from HF\nresp =\n  Req.get!(\n    \"https://huggingface.co/latent-consistency/lcm-lora-sdv1-5/resolve/main/pytorch_lora_weights.safetensors?download=true\"\n  )\n\nparam_data = resp.body\n{config, lcm_lora_params} = Lorax.Lcm.load!(param_data)\n\n# Option #2, Load Locally\n# {config, lcm_lora_params} =\n#   Lorax.Lcm.read!(\"/Users/[user]/.../pytorch_lora_weights.safetensors\")\n\n# Axon expects one map containing all the layer names -> tensors. \n# We'll merge the LCM params with the original SD params.\nlcm_unet = %{\n  unet\n  | model: Lorax.inject(unet.model, config),\n    params: Map.merge(lcm_lora_params, unet.params)\n}\n\nlcm_scheduler = %Bumblebee.Diffusion.LcmScheduler{}\n```","ref":"lcm_lora_realtime.html#load-sd1-5-lcm-lora-and-lcm-scheduler"},{"type":"extras","title":"Create serving - Stable Diffusion w/ LCM (Realtime)","doc":"```elixir\nserving =\n  Bumblebee.Diffusion.StableDiffusion.text_to_image(clip, lcm_unet, vae, tokenizer, lcm_scheduler,\n    num_steps: 4,\n    num_images_per_prompt: 1,\n    safety_checker_featurizer: featurizer,\n    guidance_scale: 1.0,\n    compile: [batch_size: 1, sequence_length: 60],\n    defn_options: [compiler: EXLA]\n  )\n```\n\n#","ref":"lcm_lora_realtime.html#create-serving"},{"type":"extras","title":"Realtime LCM LoRA - Stable Diffusion w/ LCM (Realtime)","doc":"Run all the following cells and begin typing in the prompt below. The Kino frame beneath the prompt should reflect new images. Debounce value can be changed to control how fast the images should refresh\n\nThis has been tested on the NVIDIA A10G and above.\nMake sure you use a high end graphics card for this or else the frame below will be blank for quite a long time.\n\n```elixir\ninput = Kino.Input.textarea(\"Prompt\", debounce: 150)\n```\n\n```elixir\n# Images will be rendered in this frame\nframe = Kino.Frame.new()\n```\n\n```elixir\nKino.listen(input, fn %{value: prompt} ->\n  # Generate the image and put it in the frame\n  output = Nx.Serving.run(serving, %{prompt: prompt})\n  [result] = output.results\n  image = Kino.Image.new(result.image)\n  Kino.Frame.render(frame, image)\nend)\n```","ref":"lcm_lora_realtime.html#realtime-lcm-lora"}],"content_type":"text/markdown"}